static char rcsid[] = "@(#)$Id: builtin++.c,v 2.16 2012-10-14 14:38:41 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 2.16 $   $State: Exp $
 *
 *  Modified by: Kari Hurtta <hurtta+elm@siilo.FMI.FI>
 *                       (was hurtta+elm@posti.FMI.FI, hurtta+elm@ozone.FMI.FI)
 *           or  Kari Hurtta <elm@elmme-mailer.org>
 *
 *  Initially written by: Michael Elkins <elkins@aero.org>, 1995
 *****************************************************************************/

#include "def_elm.h"
#include "s_elm.h"

DEBUG_VAR(Debug,__FILE__,"ui");

#include <errno.h>
#ifndef ANSI_C
extern int errno;
#endif         
extern int tabspacing;


static int search_it P_((struct stringbuffer * bout, int start_idx,
			 struct string * search_pattern));

static int search_it(bout,start_idx,search_pattern)
     struct stringbuffer * bout; 
     int start_idx;
     struct string * search_pattern;
{
    int pg_flags;
    int ret = 0;
    int idx = start_idx;

    while (idx < linecount_stringbuffer(bout) && !ret) {
	struct string * buffer = get_line_from_stringbuffer(bout,idx,&pg_flags);

	if (find_pattern_from_string(buffer,search_pattern,1))
	    ret = idx;
	
	idx++;
	free_string(&buffer);
    }
    DPRINT(Debug,3,
	   (&Debug,"Builtin pager: Search from line %d %s (%d)\n",
	    start_idx, 
	    ret ? "SUCCEED" : "FAILED",
	    ret));
    return ret;
}

int builtinplusplus (bout,pager_page, is_helper, pager_cmds)
     struct stringbuffer * bout;
     struct pager_page * pager_page;
     int is_helper;
     struct elm_commands *pager_cmds;
{   
    int idx = 0;
    int is_end = 0;

    /* Current line for outputting and position on it ... */
    struct string * buffer = NULL;
    int                ret = '\0';
    int             pg_flags = 0;
    int X = 0;
    struct string * search_pattern = NULL;        /* Last seach string */
   
    clear_error();
    
    for (;;) {
	int idx_top;             /* line number of top of display  */
	int ch;                  /* command key read */
	int search_matches = 0;  /* First idx on page where search matches */
	int lines;               /* Number of visible lines printed
				    so far */
	int FF_seen;            /* If page is breaken because of FF */
	int LINES, COLUMNS;

 new_page:
    resize_mark:
	idx_top = idx;       /* line number of top of display  */
	lines = 0;           /* Number of visible lines printed
				so far */
        FF_seen = 0;

	menu_ClearScreen(pager_page->root);

    cr_restart:                  /* so that we can handle CR specially --
				    ie. don't reset page variables 
				 */

	menu_get_sizes(pager_page->root, &LINES, &COLUMNS);

    

	DPRINT(Debug,5,(&Debug, "builtin++ top  idx = %d\n",idx));

	while ((idx < linecount_stringbuffer(bout)) &&
	       lines < LINES-1 && !FF_seen) {
	    
	    int chars = 0; /* Number of characters on current line */
	    int mayclear = 1;
	    int buffer_len = -1;
	    int visible_len;

	    int this_flags = 0;
	

	    if (menu_resized(pager_page->root)) {
		menu_get_sizes(pager_page->root, &LINES, &COLUMNS);

		if (buffer)
		    free_string(&buffer);
	    
		idx= idx_top;
		goto new_page;
	    }

	    if (menu_need_redraw(pager_page->root)) {

		if (buffer)
		    free_string(&buffer);

		idx= idx_top;
		goto new_page;

	    }
    
	    DPRINT(Debug,9,(&Debug, 
			    "builtin++ before line  idx = %d X=%d\n",
			    idx,X));

	    if (!buffer || X >= (buffer_len = string_len(buffer))) {
		/* caller of get_line_from_stringbuffer is expected to 
		   free_string () resulting buffer */

		if (buffer) {
		    if (X >= buffer_len)
			idx++;
		    if (idx >= linecount_stringbuffer(bout))
			break;
		    free_string (&buffer);
		}
		X = 0;

		/* There is no pending data to output, so we need to grab 
		 * another line.
		 */

		buffer = get_line_from_stringbuffer(bout,idx,&pg_flags);		
		buffer_len = string_len(buffer);


		DPRINT(Debug,9,(&Debug, 
				"builtin++ new line -- len = %d\n",buffer_len));
	    }

	    /* This is the part of the code that actually displays on 
	     * the screen 
	     */
	    chars = 0;
	    while (X < buffer_len) {
		uint16 ch;
		int len = 0;   /* Length for next clip ... */
		struct string * data = NULL;

		/* If no space for next character */
		if (chars >= COLUMNS)
		    break;
		
		ch = give_unicode_from_string(buffer,X);
		if (ch <= 31 || ch == 0x007F /* ^? */ ) {
		    
		    /* Make sure that carbage from previous
		       output is not left */
		    if (mayclear) {
			menu_CleartoEOLN(pager_page->root);
			mayclear = 0;
		    }

		    switch(ch) {
		    case 0x0009:  /* HT */
			menu_Writechar(pager_page->root,'\t');
			chars = ((chars / get_tabspacing() ) +1) * 
			    get_tabspacing(); 
			break;
		    case 0x000C:   /* FF */
		    case 0x000B:   /* VT */
			FF_seen = 1;
			/* FALLTHRU */
		    case 0x000A:   /* LF */
			X++;
			goto out;

		    case 0x007F:
			if (chars >= COLUMNS - 1)
			    goto wrap_control;

			menu_StartXX(pager_page->root,pg_REVERSE);
			menu_Writechar(pager_page->root,'^');			
			menu_Writechar(pager_page->root,'?');			
			menu_EndXX(pager_page->root,pg_REVERSE);

			chars += 2;
			break;

		    default:
			if (chars >= COLUMNS - 1)
			    goto wrap_control;

			/* Output character set is assumed to be ASCII
			   compatible on here: */

			menu_StartXX(pager_page->root,pg_REVERSE);
			menu_Writechar(pager_page->root,'^');			
			menu_Writechar(pager_page->root,ch + 64);
			menu_EndXX(pager_page->root,pg_REVERSE);

			chars += 2;
			break;
		    }

		    X++;
		    continue;
		}

		if (search_matches &&
		    search_matches == idx)
		    this_flags = pg_STANDOUT;
		else
		    this_flags = pg_flags;

		switch(ch) {
		case UNICODE_BAD_CHAR:
		    if (chars >= COLUMNS)
			    goto wrap_control;

		    menu_StartXX(pager_page->root,pg_REVERSE);
		    menu_Writechar(pager_page->root,'?');
		    menu_EndXX(pager_page->root,pg_REVERSE);
		    chars ++;
		    X++;
		    break;

		case UNICODE_NO_BREAK_SPACE:  /* Not really work */
		    if (chars >= COLUMNS)
			goto wrap_control;
		    
		    if (this_flags)
			menu_StartXX(pager_page->root,this_flags);
		    menu_Writechar(pager_page->root,' ');
		    if (this_flags)
			menu_EndXX(pager_page->root,this_flags);
		    X++;
		    break;

		case UNICODE_SOFT_HYPHEN:
		    if (chars >= COLUMNS)
			goto wrap_control;

		    if (chars > COLUMNS-2) {
			DPRINT(Debug,9,(&Debug, 
					"builtin++: wrap on soft hyphen\n"));

			if (this_flags)
			    menu_StartXX(pager_page->root,this_flags);
			menu_Writechar(pager_page->root,'-');
			if (this_flags)
			    menu_EndXX(pager_page->root,this_flags);
			X++;	
			goto out;
		    }
		    X++;

		    break;

		default:
		    if (0 == unicode_ch(ch,UOP_noctrl)) {
			if (chars >= COLUMNS - 6)
			    goto wrap_control;
			menu_StartXX(pager_page->root,pg_REVERSE);

			menu_Write_to_screen(pager_page->root,
					     FRM("^u%04X"),ch);
			
			menu_EndXX(pager_page->root,pg_REVERSE);
			chars += 6;
			X++;
			break;
		    }
		}

		/* wrapping after control */
		if (chars >= COLUMNS) {
		wrap_control:
		    
		    DPRINT(Debug,9,(&Debug, 
				    "builtin++ --- No space for next character on line\n"));

		    if (chars < 1) {
			DPRINT(Debug,9,(&Debug, 
					"builtin++: bad ...\n"));
			goto read_key;
		    }

		    goto out;
		}



		DPRINT(Debug,9,(&Debug, 
				"builtin++ after ctrl X=%d (chars = %d)\n",
				X,chars));
	
		/* calculate clip len ... */
		while (X + len     < buffer_len &&
		       chars + len < COLUMNS) {
		    ch = give_unicode_from_string(buffer,X+len);

		    if (ch <= 31 || ch == 0x007F /* ^? */)
			break;

		    switch(ch) {
		    case UNICODE_BAD_CHAR:
		    case UNICODE_NO_BREAK_SPACE:
		    case UNICODE_SOFT_HYPHEN:
			goto text_found;
		    default:
			if (0 == unicode_ch(ch,UOP_noctrl))
			    goto text_found;
			break;
		    }

		    len++;
		}
		
	    text_found:

		DPRINT(Debug,9,(&Debug, 
				"builtin++ after text  X=%d ( len = %d)\n",
				X,len));
		
		if (len < 1)
		    continue;    /* next character was also control char */

		/* clip_from_string updates X */
		data = curses_printable_clip(buffer,&X,len,&visible_len,
					     COLUMNS - chars);

		if (data) {

		    chars += visible_len;   
		    
		    DPRINT(Debug,9,(&Debug, 
				    "builtin++ after clip  idx = %d X=%d (len = %d, chars=%d  visible_len=%d)\n",
				    idx,X,len,chars, visible_len));
		    
		    if (this_flags)
			menu_StartXX(pager_page->root,this_flags);

		    menu_Write_to_screen(pager_page->root,FRM("%S"),data);

		    if (this_flags)
			menu_EndXX(pager_page->root,this_flags);

		    free_string(&data);
		    
		    if (visible_len < 1) {
			DPRINT(Debug,9,(&Debug, 
					"builtin++ --- No space for next character on line\n"));

			if (chars < 1) {
			    DPRINT(Debug,9,(&Debug, 
					    "builtin++: bad ...\n"));
			    goto read_key;
			}

			goto out;
		    }

		} else {
		    /* Treat next character as control character */

		    /* Output character set is assumed to be ASCII
		       compatible on here: */
		    menu_StartXX(pager_page->root,pg_REVERSE);
		    menu_Writechar(pager_page->root,'?');		    
		    menu_EndXX(pager_page->root,pg_REVERSE);

		    chars += 1;
		    X++;
		}
	    }



	out:
	    /* Make sure that carbage from previous
	       output is not left */
	    if (mayclear) {
		menu_CleartoEOLN(pager_page->root);
		mayclear = 0;
	    }

	    menu_Writechar(pager_page->root,'\r');		    
	    menu_Writechar(pager_page->root,'\n');		    
	    lines++;
	}

	menu_CleartoEOS(pager_page->root);
	if (lines < LINES-1 && FF_seen) {
	    menu_StartXX(pager_page->root,pg_BOLD);
	    menu_PutLineX (pager_page->root,
			   lines, 0, 
			   CATGETS(elm_msg_cat, ElmSet, ElmBuiltinNewPage,
				   "~ (new page) "));
	    menu_EndXX(pager_page->root,pg_BOLD);  
	    menu_CleartoEOLN(pager_page->root);
	    menu_Writechar(pager_page->root,'\r');		    
	    menu_Writechar(pager_page->root,'\n');		    

	    lines++;
	}
	while (lines < LINES-1) {
	    menu_Writechar(pager_page->root,'~');
	    menu_CleartoEOLN(pager_page->root); 
	    menu_Writechar(pager_page->root,'\r');		    
	    menu_Writechar(pager_page->root,'\n');		    
	    
	    lines++;
	}

    reprompt:
	menu_StartXX(pager_page->root,pg_STANDOUT);
	
	if (idx >= linecount_stringbuffer(bout)) {
	    if (is_helper)
		menu_PutLineX (pager_page->root,LINES-1, 0, 
			       CATGETS(elm_msg_cat, ElmSet, ElmBuiltinHelpEnd,
				       "Press 'q' or <return> to return:"));

	    else
		menu_PutLineX (pager_page->root,LINES-1, 0, 
			       CATGETS(elm_msg_cat, ElmSet, ElmBuiltinCommandi,
				       "Command ('i' to return to index):"));
	    is_end = 1;
	}
	else {
	    int relative;
	    enum user_level_v ul = 
		give_dt_enumerate_as_int(&user_level);

	    if (linecount_stringbuffer(bout) < 1)
		relative = 100;
	    else
		relative = 100 * idx / linecount_stringbuffer(bout);
	    
	    if (is_helper) {
		menu_PutLineX(pager_page->root,
			      LINES-1, 0,CATGETS(elm_msg_cat, ElmSet, ElmBuiltHelpMore,
						 "Press <space> to continue, 'q' to return:"));
	    } else if (ul == user_level_beginner)
		menu_PutLineX(pager_page->root,LINES-1, 0, 
			      CATGETS(elm_msg_cat, ElmSet, ElmBuiltInMore0,
				      "MORE (line %d-%d/%d, you've seen %d%%). Press <space> for more:"), 
			      idx_top+1,idx,linecount_stringbuffer(bout),
			      relative);
	    else
		menu_PutLineX(pager_page->root,
			      LINES-1, 0, 
			      CATGETS(elm_msg_cat, ElmSet, ElmBuiltInMore,
				      "MORE (line %d-%d/%d, you've seen %d%%):"), 
			      idx_top+1,idx,linecount_stringbuffer(bout),
			      relative);
	    is_end = 0;
	}
	menu_EndXX(pager_page->root,pg_STANDOUT);
	menu_CleartoEOS(pager_page->root);

    read_key:
	ch = menu_ReadCh(pager_page->root, REDRAW_MARK|READCH_CURSOR|READCH_resize|
			 READCH_sig_char);

	if (ch != '/') {
	    if (search_pattern) {
		free_string(&search_pattern);
	    }
	}

	switch (ch) {
	case TERMCH_interrupt_char:
	    ret = 0;
	    goto finish;

	case RESIZE_MARK:
	    DPRINT(Debug,4, (&Debug, "    ... resizing\n"));
	    idx = idx_top;
	    goto resize_mark;

	case ' ':
	case '+':
	case PAGEDOWN_MARK:
	    if (is_end) {
		ret = ' ';
		goto finish;
	    }
	break;
	case HOME_MARK:
	case '^':      
	    idx = 0;
	    /* Avoid incrementing of idx: */
	    if (buffer)    
		free_string (&buffer);
	    break;
	case 'G':
	    /* Not completely correct if multiline lines */
	    idx = linecount_stringbuffer(bout)-LINES;
	    if (idx < 0)
		idx = 0;
	    lines = 0;
	    /* Avoid incrementing of idx: */
	    if (buffer)    
		free_string (&buffer);
	    menu_ClearScreen(pager_page->root);
	    break;
	case FIND_MARK:
	case '/':
	    menu_ClearLine (pager_page->root,LINES-1);
	    {
		int code;
		
		code = optionally_enter2(pager_page->root,
					 &search_pattern,
					 LINES-1, 0, 
					 OE_REDRAW_MARK|
					 OE_SIG_CHAR /* Ctrl-C */, 
					 CATGETS(elm_msg_cat, ElmSet, 
						 ElmBuiltinSearch,
						 "Search: "));

		if (REDRAW_MARK == code)
		    goto redraw;
		if (-1 == code) /* Ctrl-C */
		    goto reprompt;
		if (code != 0)
		    goto quit;
		if (search_pattern && string_len(search_pattern)) {
		    int start_idx;
		    /* Continue search from previous match */
		    if (!search_matches)
			start_idx = idx_top;
		    else if (search_matches < linecount_stringbuffer(bout))
			start_idx = search_matches+1;
		    else
			start_idx = idx_top;
		    search_matches = search_it(bout,start_idx,
					       search_pattern);
		    if (search_matches) {
			if (search_matches > idx_top && 
			    search_matches < idx_top + LINES -4)
			    idx = idx_top;
			else
			    idx = search_matches - LINES / 2;
			if (idx < 0)
			    idx = 0;
			/* Avoid incrementing of idx: */
			if (buffer)    
			    free_string (&buffer);
			goto new_page; /* and draw match */
		    } else {
			int relative;
			enum user_level_v ul = 
			    give_dt_enumerate_as_int(&user_level);

			if (linecount_stringbuffer(bout) < 1)
			    relative = 100;
			else
			    relative = 100 * idx / linecount_stringbuffer(bout);
			
			menu_ClearLine (pager_page->root,LINES-1);
			menu_StartXX(pager_page->root,pg_STANDOUT);
			if (ul == user_level_beginner)
			    menu_PutLineX (pager_page->root,LINES-1, 0, 
					   CATGETS(elm_msg_cat, ElmSet, ElmBuiltInNotFound0,
						   "NOT FOUND! (line %d-%d/%d, you've seen %d%%). Press <space> for more:"), 
				      idx_top+1,idx,linecount_stringbuffer(bout),
				      relative);
		    else
			menu_PutLineX (pager_page->root,LINES-1, 0, 
				       CATGETS(elm_msg_cat, ElmSet, 
					       ElmBuiltInNotFound,
					       "NOT FOUND! (line %d-%d/%d, you've seen %d%%):"), 
				       idx_top+1,idx,linecount_stringbuffer(bout),
				  relative);		    
			menu_EndXX(pager_page->root,pg_STANDOUT);
			goto read_key;
		    }
		} 
		goto reprompt;
	    }	    

	case PAGEUP_MARK:
	case '-':
	    idx = idx_top - LINES;
	    if (idx < 0)
		idx = 0;
	    if (buffer)
		free_string(&buffer);
	    break;
	case ctrl('P'):
	    idx = idx_top -1;
	    if (idx < 0)
		idx = 0;
	    if (buffer)
		free_string(&buffer);
	    break;
	case ctrl('D'):
	    if (is_end) 
		goto reprompt; /* Can't go down */
	    lines = LINES / 2;
	    menu_Writechar(pager_page->root,'\r');		    

	    menu_CleartoEOLN(pager_page->root);	    
	    idx_top += LINES / 2;
	    goto cr_restart; /* Hanlde Ctrl-D specially */
	    	   	    
	case ctrl('U'):
	    idx = idx_top - LINES / 2;
	    if (idx < 0)
		idx = 0;
	    if (buffer)
		free_string(&buffer);
	    break;
	case HELP_MARK:
	case '?': {

	    if (is_helper < 1) {
		struct elm_commands *cmds = 
		    give_builtinpager_commands();

		int ch;

	        if (pager_cmds) {
		    struct elm_commands *all_cmds =
			give_merged_commands(cmds,pager_cmds,
					     pg_BOLD,0);
		    
		    struct stringbuffer * b =
			give_commands_help(all_cmds);

		    ch = builtinplusplus(b,pager_page,is_helper+1,NULL);

		    free_stringbuffer(&b);
		    free_commands(&all_cmds);
		} else {
		    struct stringbuffer * b =
			give_commands_help(cmds);

		    ch = builtinplusplus(b,pager_page,is_helper+1,NULL);

		    free_stringbuffer(&b);
		}
		free_commands(&cmds);

		if (EOF == ch) {
		    ret = EOF;
		    goto finish;
		}
	    }	   
	}
	    /* FALLTHRU */
	redraw:
	case REDRAW_MARK:
	case ctrl('L'): 
	    idx = idx_top;
	    if (buffer)
		free_string(&buffer);
	    break;
	case '\n':
	case '\r':
	    if (FF_seen)
		break;   /* New page */
	    {
		int x, y;
		
		GetXYLocation (&x, &y);
		ClearLine (x);
	    } 

	    if (is_helper && is_end) {
		ret = 0;
		goto finish;
	    }

	    lines = LINES - 2;
	    idx_top++;
	    goto cr_restart; /* Hanlde CR specially */

	quit:
	case EOF:
	default:
	    ret =  (ch == 'q' || ch == 'x' ? 0 : ch);
	    goto finish;
	}
    }

 finish:

    if (buffer)
	free_string (&buffer);
    DPRINT(Debug,9,(&Debug, 
		    "builtin++ returning %c\n",ret));
    return ret;
}

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
